import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here,
                       'VERSION.md')) as version_file:
    __version__ = version_file.read().strip()

default_app_config = 'template_engines.apps.TemplateEnginesConfig'



import io
import os
import zipfile
from zipfile import BadZipFile

from bs4 import BeautifulSoup
from django.core.files.storage import default_storage
from django.template.exceptions import TemplateDoesNotExist
from django.template.backends.django import DjangoTemplates
from django.utils.functional import cached_property

from template_engines.utils import clean_tags


class AbstractTemplate:
    """
    Gives the architecture of a basic template.
    """

    def __init__(self, template):
        self.template = template

    def clean(self, data):
        """
        Method for cleaning rendered data.
        """
        raise NotImplementedError()

    def render(self, context=None, request=None):
        """
        Fills a template with the context obtained by combining the `context` and` request`
        parameters and returns a file as a byte object.
        """
        raise NotImplementedError()


class BaseEngine(DjangoTemplates):
    app_dirname = None
    sub_dirname = None
    template_class = None

    @cached_property
    def template_dirs(self):
        t_dirs = super().template_dirs
        if self.sub_dirname:
            t_dirs += tuple([os.path.join(p, self.sub_dirname) for p in t_dirs])
        return t_dirs


class ZipAbstractEngine(BaseEngine):
    """
    Gives the architecture of a basic zip base template engine.

    Can be specified:

    * ``app_dirname``, the folder name which contains the templates in application directories,
    * ``sub_dirname``, the folder name of the subdirectory in the templates directory,
    * ``template_class``, your own template class with a ``render`` method,
    * ``zip_root_file``, the file to fill.
    """
    zip_root_files = None

    def get_template_content(self, filename):
        """
        Returns the contents of a template before modification, as a string.
        """
        try:
            with default_storage.open(filename, 'rb') as template_file:
                template_buffer = io.BytesIO(template_file.read())

                with zipfile.ZipFile(template_buffer, 'r') as zip_file:
                    soup = BeautifulSoup('', 'html.parser')
                    global_tag = soup.new_tag("global-merged")

                    for file_to_merge in self.zip_root_files:
                        name, _ = os.path.splitext(file_to_merge)
                        content = zip_file.read(file_to_merge).decode()
                        content_soup = BeautifulSoup(content, 'xml')
                        tag_file = content_soup.new_tag('{0}-merged'.format(name.replace('/', '-')))
                        tag_file.append(content_soup)
                        global_tag.append(tag_file)

                    soup.append(global_tag)
                    return str(soup)
        except (KeyError, BadZipFile):
            raise TemplateDoesNotExist('Bad format.')

    def from_string(self, template_code, **kwargs):
        return self.template_class(self.engine.from_string(template_code), **kwargs)

    def get_template_path(self, filename):
        """
        Check if a template named ``template_name`` can be found in a list of directories. Returns
        the path if the file exists or raises ``TemplateDoesNotExist`` otherwise.
        """
        if default_storage.exists(filename):
            return filename
        for directory in self.template_dirs:
            abstract_path = default_storage.generate_filename(os.path.join(directory, filename))
            if default_storage.exists(abstract_path):
                return abstract_path
        raise TemplateDoesNotExist(f'Unknown: {filename}')

    def clean_content(self, content):
        return clean_tags(content)

    def get_template(self, template_name):
        template_path = self.get_template_path(template_name)
        content = self.get_template_content(template_path)
        content = self.clean_content(content)
        return self.from_string(content, template_path=template_path)



from pathlib import Path

import weasyprint
from django.template.backends.utils import csrf_input_lazy, csrf_token_lazy
from django.template.context import make_context
from django.template.exceptions import TemplateDoesNotExist

from template_engines import settings as app_settings, settings
from template_engines.backends import AbstractTemplate, BaseEngine


class WeasyprintTemplate(AbstractTemplate):
    def get_base_url(self, request):
        """
        Determine base URL to fetch static and media files from `WEASYPRINT_BASEURL` or
        fall back to using the root path of the URL used in the request.
        """
        return getattr(
            settings, 'WEASYPRINT_BASEURL',
            request.build_absolute_uri('/') if request else '/'
        )

    def render(self, context=None, request=None):
        base_url = self.get_base_url(request)
        if context is None:
            context = {}
        if request is not None:
            context['request'] = request
            context['csrf_input'] = csrf_input_lazy(request)
            context['csrf_token'] = csrf_token_lazy(request)

        html = weasyprint.HTML(
            string=self.template.render(make_context(context)),
            base_url=base_url,
        )
        html.render()
        return html.write_pdf()


class WeasyprintEngine(BaseEngine):
    sub_dirname = app_settings.WEASYPRINT_ENGINE_SUB_DIRNAME
    app_dirname = app_settings.WEASYPRINT_ENGINE_APP_DIRNAME
    template_class = WeasyprintTemplate

    def get_template(self, template_name):
        template_path = Path(template_name)
        if template_path.suffix.lower() != '.pdf':
            raise TemplateDoesNotExist('This is not a template PDF file')
        return self.template_class(self.engine.get_template(template_name))



import base64
import logging
import random
import re
import secrets

import requests
from bs4 import BeautifulSoup
from django import template
from django.utils.safestring import mark_safe

from template_engines.utils.odt import ODT_IMAGE
from .utils import get_extension_picture, parse_tag, resize, get_image_infos_from_uri

register = template.Library()

logger = logging.getLogger(__name__)


def parse_p(soup):
    """ Replace p tags with text:p """
    paragraphs = soup.find_all("p")

    for p_tag in paragraphs:
        p_tag.name = 'text:p'
    return soup


def parse_span(soup):
    """ Replace span tags with text:span """
    spans = soup.find_all("span")

    for span_tag in spans:
        span_tag.name = 'text:span'
    return soup


def parse_italic(soup):
    """ Replace i tags with text:span with automatic style """
    italic_tags = soup.find_all("em")
    for i_tag in italic_tags:
        i_tag.name = 'text:span'
        i_tag.attrs['text:style-name'] = "ITALIC"
    return soup


def parse_strong(soup):
    """ Replace strong tags with text:span with automatic style """
    strong_tags = soup.find_all("strong")
    for s_tag in strong_tags:
        s_tag.name = 'text:span'
        s_tag.attrs['text:style-name'] = "BOLD"
    return soup


def parse_underline(soup):
    """ Replace u tags with text:span with automatic style """
    u_tags = soup.find_all("u")
    for u_tag in u_tags:
        u_tag.name = 'text:span'
        u_tag.attrs['text:style-name'] = "UNDERLINE"
    return soup


def fix_li(soup, tag):
    tag.name = 'text:list-item'
    contents = ''
    for e in tag.contents:
        # get tag content formatted (keep nested tags)
        contents = f'{contents}{e}'
    tag.string = ''
    content = soup.new_tag('text:p')
    content.attrs['text:style-name'] = "Standard"
    content.append(BeautifulSoup(contents, 'html.parser'))
    tag.append(content)


def parse_ul(soup):
    """ Replace ul / li tags text:list and text:list-item """
    ul_tags = soup.find_all("ul")

    for ul_tag in ul_tags:
        ul_tag.name = 'text:list'
        ul_tag.attrs['xml:id'] = f'list{str(random.randint(100000000000000000, 900000000000000000))}'
        ul_tag.attrs['text:style-name'] = "L1"
        li_tags = ul_tag.findChildren(recursive=False)

        for li in li_tags:
            fix_li(soup, li)

    return soup


def parse_ol(soup):
    """ Replace ol / li tags text:list and text:list-item """
    ol_tags = soup.find_all("ol")
    for ol_tag in ol_tags:
        ol_tag.name = 'text:list'
        ol_tag.attrs['xml:id'] = f'list{str(random.randint(100000000000000000, 900000000000000000))}'
        ol_tag.attrs['text:style-name'] = "L2"
        ol_tag.attrs['text:level'] = "1"
        li_tags = ol_tag.findChildren(recursive=False)

        for li in li_tags:
            fix_li(soup, li)
    return soup


def parse_a(soup):
    # replace a
    a_tags = soup.find_all("a")
    for a_tag in a_tags:
        a_tag.name = 'text:a'
        new_attrs = {
            'xlink:type': 'simple',
            'xlink:href': a_tag.attrs['href']
        }
        a_tag.attrs = new_attrs

    return soup


def parse_h(soup):
    # replace h*
    h_tags = soup.find_all(re.compile("^h[0-3]"))
    for h_tag in h_tags:
        num = h_tag.name[1]
        h_tag.name = 'text:h'
        new_attrs = {
            'text:outline-level': num
        }
        h_tag.attrs = new_attrs
    return soup


def parse_br(soup):
    # replace br
    br_tags = soup.find_all("br")
    for br_tag in br_tags:
        br_tag.name = 'text:line-break'
    return soup


def parse_img(soup):
    """ Replace img tags with text:p """
    imgs = soup.find_all("img")
    # TODO: if src starts with http / https, download file and use local path in odt
    for img in imgs:
        img.name = 'draw:frame'
        src = img.attrs.pop('src')
        content = soup.new_tag('draw:image')
        size, dimensions, mime_type = get_image_infos_from_uri(src)
        content.attrs = {
            'xlink:href': src,
            'xlink:type': "simple",
            'xlink:show': "embed",
            'xlink:actuate': "onload",
            'draw:mime-type': mime_type
        }

        width, height = dimensions
        if width and width > 500:
            # if sized and sized > 500, it will not fit to page width
            ratio = width / height
            # keep ratio
            width = 500
            height = int(width / ratio)
        img.attrs = {
            'draw:style-name': "fr1",
            'text:anchor-type': "as-char",
            'svg:width': f"{width}px",
            'svg:height': f"{height}px",
            'draw:z-index': "37",
        }
        img.append(content)

    return soup


@register.filter()
def from_html(value, is_safe=True):
    """ Convert HTML from rte fields to odt compatible format """
    soup = BeautifulSoup(value, "html.parser")
    soup = parse_span(soup)
    soup = parse_p(soup)
    soup = parse_strong(soup)
    soup = parse_italic(soup)
    soup = parse_underline(soup)
    soup = parse_ul(soup)
    soup = parse_ol(soup)
    soup = parse_a(soup)
    soup = parse_h(soup)
    soup = parse_br(soup)
    soup = parse_img(soup)
    return mark_safe(str(soup))


class ImageLoaderNodeURL(template.Node):
    def __init__(self, url, data=None, request=None, max_width=None,
                 max_height=None, anchor=None):
        # saves the passed obj parameter for later use
        # this is a template.Variable, because that way it can be resolved
        # against the current context in the render method
        self.url = url
        self.data = data
        self.request = request
        self.max_width = max_width
        self.max_height = max_height
        self.anchor = anchor

    def render(self, context):
        url, type_request, max_width, max_height, anchor, data = self.get_value_context(context)
        name = secrets.token_hex(15)
        response = self.get_content_url(url, type_request or "get", data)
        if not response:
            return ""
        width, height = resize(response.content, max_width, max_height, odt=True)
        context.setdefault('images', {})
        picture = response.content
        extension = get_extension_picture(picture)
        full_name = '{}.{}'.format(name, extension)
        context['images'].update({full_name: picture})
        return mark_safe(ODT_IMAGE.format(full_name, width, height,
                                          anchor or "paragraph", f"image/{extension.lower()}"))

    def get_value_context(self, context):
        final_url = self.url.resolve(context)
        final_request = "get" if not self.request else self.request.resolve(context)
        final_max_width = None if not self.max_width else self.max_width.resolve(context)
        final_max_height = None if not self.max_height else self.max_height.resolve(context)
        final_anchor = "paragraph" if not self.anchor else self.anchor.resolve(context)
        final_data = "" if not self.data else self.data.resolve(context)
        return final_url, final_request, final_max_width, final_max_height, final_anchor, final_data

    def get_content_url(self, url, type_request, data):
        try:
            response = getattr(requests, type_request.lower())(url, data=data)
        except requests.exceptions.ConnectionError:
            logger.warning("Connection Error, check the url given")
            return
        except AttributeError:
            logger.warning("Type of request specified not allowed")
            return
        if response.status_code != 200:
            logger.warning("The picture is not accessible (Error: %s)" % response.status_code)
            return
        return response


@register.tag
def image_url_loader(parser, token):
    """
    Replace a tag by an image from the url you specified.
    The necessary key is url
    - url : Url where you want to get your picture
    Other keys : data, max_width, max_height, request
    - data : Use it only with post request
    - max_width : Width of the picture rendered
    - max_heigth : Height of the picture rendered
    - request : Type of request, post or get. Get by default.
    - anchor : Type of anchor, paragraph, as-char, char, frame, page
    """
    tag_name, args, kwargs = parse_tag(token, parser)
    usage = '{{% {tag_name} [url] max_width="5000px" max_height="5000px" ' \
            'request="GET" data="{{"data": "example"}}" anchor="as-char" %}}'.format(tag_name=tag_name)
    if len(args) > 1 or not all(
            key in ['max_width', 'max_height', 'request', 'data', 'anchor'] for key in kwargs.keys()):
        raise template.TemplateSyntaxError("Usage: %s" % usage)
    return ImageLoaderNodeURL(*args, **kwargs)


class ImageLoaderNode(template.Node):
    def __init__(self, instance, max_width=None, max_height=None, anchor=None):
        # saves the passed obj parameter for later use
        # this is a template.Variable, because that way it can be resolved
        # against the current context in the render method
        self.object = instance
        self.max_width = max_width
        self.max_height = max_height
        self.anchor = anchor

    def base64_to_binary(self, picture):
        if isinstance(picture, str) and 'base64' in picture:
            picture = base64.b64decode(picture.split(';base64,')[1])
        return picture

    def render(self, context):
        # Evaluate the arguments in the current context
        name = self.object
        picture, max_width, max_height, anchor = self.get_value_context(context)
        picture = self.base64_to_binary(picture)
        if not picture or not isinstance(picture, bytes):
            logger.warning("{object} is not a valid picture".format(object=name))
            return ""

        name = secrets.token_hex(15)
        width, height = resize(picture, max_width, max_height, odt=True)
        context.setdefault('images', {})
        extension = get_extension_picture(picture)
        full_name = '{}.{}'.format(name, extension)
        context['images'].update({full_name: picture})
        return mark_safe(ODT_IMAGE.format(full_name, width, height, anchor or "paragraph",
                                          f"image/{extension.lower()}"))

    def get_value_context(self, context):
        final_object = self.object.resolve(context)
        final_max_width = None if not self.max_width else self.max_width.resolve(context)
        final_max_height = None if not self.max_height else self.max_height.resolve(context)
        final_anchor = "paragraph" if not self.anchor else self.anchor.resolve(context)
        return final_object, final_max_width, final_max_height, final_anchor


@register.tag
def image_loader(parser, token):
    """
    Replace a tag by an image you specified.
    You must add an entry to the ``context`` var that is a dict with at least a ``content`` key
    whose value is a byte object. You can also specify ``max_width`` and ``max_height``.
    The necessary key is image
    - image : content of your picture in binary or base64
    Other keys : max_width, max_height, anchor
    - max_width : Width of the picture rendered
    - max_height : Height of the picture rendered
    - anchor : Type of anchor, paragraph, as-char, char, frame, page
    """
    tag_name, args, kwargs = parse_tag(token, parser)
    usage = '{{% {tag_name} [image] max_width="5000px" max_height="5000px" anchor="as-char" %}}'.format(
        tag_name=tag_name)
    if len(args) > 1 or not all(key in ['max_width', 'max_height', 'anchor']
                                for key in kwargs.keys()):
        raise template.TemplateSyntaxError("Usage: %s" % usage)
    return ImageLoaderNode(*args, **kwargs)



from bs4 import BeautifulSoup
import os
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

ODT_IMAGE = (
    '<draw:frame draw:name="{0}" svg:width="{1}" svg:height="{2}" text:anchor-type="{3}" draw:z-index="37">'
    '<draw:image xlink:href="Pictures/{0}" xlink:type="simple" xlink:show="embed" xlink:actuate="onLoad" draw:mime-type="{4}" /></draw:frame>'
)


def add_image_in_odt_template(bfile, image, name):
    """
    Makes an image available for a odt document.

    :param bfile: the odt document.
    :type bfile: bytes

    :param image: a dictionary with at least two entries: the image's name and the image's content.
    :type image: dict
    """
    temp_file = NamedTemporaryFile()
    with open(temp_file.name, 'wb') as writer:
        writer.write(bfile)

    temp_doc = NamedTemporaryFile()
    with ZipFile(temp_file.name, 'r') as read_zip_file:
        info_list = read_zip_file.infolist()
        with ZipFile(temp_doc.name, 'w') as write_zip_file:
            # Add a relationship tag to word/_rels/document.xml.rels
            for item in info_list:
                if item.filename == 'META-INF/manifest.xml':
                    doc_relationships = read_zip_file.read(item.filename)
                    soup = BeautifulSoup(doc_relationships, features='xml')
                    manifest = soup.find('manifest:manifest')
                    file_entry = soup.new_tag('manifest:file-entry')
                    _, ext = os.path.splitext(name)
                    attributes_file_entry = {
                        'manifest:full-path': 'Pictures/{}'.format(name),
                        'manifest:media-type': 'image/{}'.format(ext)
                    }
                    file_entry.attrs = attributes_file_entry
                    manifest.append(file_entry)
                    write_zip_file.writestr(item.filename, str(soup))
                else:
                    write_zip_file.writestr(item, read_zip_file.read(item.filename))

            # Add the image in the word/media folder
            temp_image = NamedTemporaryFile()

            with open(temp_image.name, 'wb') as writer:
                writer.write(image)
            write_zip_file.write(
                temp_image.name,
                arcname='Pictures/{}'.format(name))
    with open(temp_doc.name, 'rb') as read_file:
        return read_file.read()



import re
from tempfile import NamedTemporaryFile
from zipfile import ZipFile


DOCX_RELATIONSHIP = (
    '<Relationship Id="{0}" Type="http://schemas.openxmlformats.org/officeDocument/2006/'
    + 'relationships/image" Target="media/{0}"/></Relationships>')


DOCX_PARAGRAPH_RE = re.compile(
    r'<w:r><w:rPr>([^>]*)</w:rPr><w:t>[^<]+</w:t></w:r\>')
DOCX_CHANGES = {
    '\n': '</w:t><w:br/><w:t>',
    '&lt;b&gt;': '</w:t></w:r><w:r><w:rPr>{}<w:b w:val="true"/></w:rPr><w:t>&#xA0;',
    '&lt;/b&gt;': '&#xA0;</w:t></w:r><w:r><w:rPr>{}<w:b w:val="false"/></w:rPr><w:t>',
}


def add_image_in_docx_template(bfile, image):
    """
    Makes an image available for a docx document.

    :param bfile: the docx document.
    :type bfile: bytes

    :param image: a dictionary with at least two entries: the image's name and the image's content.
    :type image: dict
    """
    temp_file = NamedTemporaryFile()
    with open(temp_file.name, 'wb') as writer:
        writer.write(bfile)

    temp_doc = NamedTemporaryFile()
    with ZipFile(temp_file.name, 'r') as read_zip_file:
        info_list = read_zip_file.infolist()
        with ZipFile(temp_doc.name, 'w') as write_zip_file:
            # Add a relationship tag to word/_rels/document.xml.rels
            for item in info_list:
                if item.filename == 'word/_rels/document.xml.rels':
                    doc_relationships = read_zip_file.read(item.filename).decode()
                    doc_relationships = re.sub(
                        '</Relationships>',
                        DOCX_RELATIONSHIP.format(image.get('name')),
                        doc_relationships)
                    write_zip_file.writestr(item.filename, doc_relationships)
                else:
                    write_zip_file.writestr(item, read_zip_file.read(item.filename))

            # Add the image in the word/media folder
            temp_image = NamedTemporaryFile()
            with open(temp_image.name, 'wb') as writer:
                writer.write(image.get('content'))
            write_zip_file.write(
                temp_image.name,
                arcname='word/media/{}'.format(image.get('name')))

    with open(temp_doc.name, 'rb') as read_file:
        return read_file.read()


TO_CHANGE_RE = re.compile(r'\n|&lt;b&gt;|&lt;/b&gt;')
